#include "config.h"
#include <libexif/exifmem.h>

#include <stdlib.h>

#ifdef __PUREC__
struct _ExifEntryPrivate { int dummy; };
struct _ExifContentPrivate { int dummy; };
struct _ExifMnoteData { int dummy; };
struct _ExifDataPrivate { int dummy; };
struct _ExifLog { int dummy; };
#endif

/*! ExifMem define a memory allocator */
struct _ExifMem {
	uint32_t ref_count;
	ExifMemAllocFunc alloc_func;
	ExifMemReallocFunc realloc_func;
	ExifMemFreeFunc free_func;
};


/*! Default memory allocation function. */
static void *EXIF_CALLBACK exif_mem_alloc_func (ExifLong ds)
{
	return calloc ((size_t) ds, 1);
}

/*! Default memory reallocation function. */
static void *EXIF_CALLBACK exif_mem_realloc_func (void *d, ExifLong ds)
{
	return realloc (d, (size_t) ds);
}

/*! Default memory free function. */
static void EXIF_CALLBACK exif_mem_free_func (void *d)
{
	free (d);
}

EXIF_API(ExifMem *) exif_mem_new (ExifMemAllocFunc alloc_func, ExifMemReallocFunc realloc_func, ExifMemFreeFunc free_func)
{
	ExifMem *mem;

	if (!alloc_func && !realloc_func) 
		return NULL;
	mem = alloc_func ? alloc_func (sizeof (ExifMem)) :
		           realloc_func (NULL, sizeof (ExifMem));
	if (!mem) return NULL;
	mem->ref_count = 1;

	mem->alloc_func   = alloc_func;
	mem->realloc_func = realloc_func;
	mem->free_func    = free_func;

	return mem;
}

EXIF_API(void) exif_mem_ref (ExifMem *mem)
{
	if (!mem) return;
	mem->ref_count++;
}

EXIF_API(void) exif_mem_unref (ExifMem *mem)
{
	if (!mem) return;
	if (!--mem->ref_count)
		exif_mem_free (mem, mem);
}

EXIF_API(void) exif_mem_free (ExifMem *mem, void *d)
{
	if (!mem) return;
	if (mem->free_func) {
		mem->free_func (d);
		return;
	}
}

EXIF_API(void *) exif_mem_alloc (ExifMem *mem, ExifLong ds)
{
	if (!mem) return NULL;
	if (mem->alloc_func || mem->realloc_func)
		return mem->alloc_func ? mem->alloc_func (ds) :
					 mem->realloc_func (NULL, ds);
	return NULL;
}

EXIF_API(void *) exif_mem_realloc (ExifMem *mem, void *d, ExifLong ds)
{
	return (mem && mem->realloc_func) ? mem->realloc_func (d, ds) : NULL;
}

EXIF_API(ExifMem *) exif_mem_new_default (void)
{
	return exif_mem_new (exif_mem_alloc_func, exif_mem_realloc_func, exif_mem_free_func);
}
